/*
 *	Simple User Authentication System -- The Getty
 *
 *	(c) 1996--1998 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License. See file COPYING in any of the GNU packages.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <unistd.h>
#include <utmp.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <syslog.h>
#include <sys/utsname.h>
#include <ctype.h>
#include <pwd.h>
#include <grp.h>
#include <sys/un.h>
#include <sys/param.h>

#include "suas.h"
#include "master.h"

/*
 *	General Routines
 */

static void threrr(char *c, ...) NO_RET;

static void
threrr(char *c, ...)
{
  va_list args;

  va_start(args, c);
  fprintf(stderr, "getty[%d]: ", getpid());
  vfprintf(stderr, c, args);
  fputc('\n', stderr);
  sleep(1);
  exit(1);
}

/*
 *	UTMP and WTMP
 */

static void
update_wtmp(struct utmp *ut)
{
  int fd;

  if ((fd = open(_PATH_WTMP, O_APPEND | O_WRONLY)) >= 0)
    {
      flock(fd, LOCK_EX | LOCK_NB);
      write(fd, ut, sizeof(ut));
      flock(fd, LOCK_UN | LOCK_NB);
      close(fd);
    }
}

static struct utmp *
find_utmp(pid_t pid)
{
  struct utmp *ut;
  static struct utmp utt;

  utmpname(_PATH_UTMP);
  setutent();
  while ((ut = getutent()) && ut->ut_pid != pid)
    ;
  if (!ut)
    {
      ut = &utt;
      bzero(ut, sizeof(utt));
      ut->ut_pid = pid;
    }
  time(&ut->ut_time);
  endutent();
  return ut;
}

static void
clean_utmp(pid_t pid, char *tty, int id)
{
  struct utmp *ut = find_utmp(pid);

  if (tty)
    {
      strcpy(ut->ut_user, "LOGIN");
      strcpy(ut->ut_line, tty);
      ut->ut_type = LOGIN_PROCESS;
      sprintf(ut->ut_id, "c%d", id);
    }
  else
    {
      bzero(&ut->ut_user, sizeof(ut->ut_user));
      bzero(&ut->ut_host, sizeof(ut->ut_host));
      ut->ut_type = DEAD_PROCESS;
      ut->ut_pid = 0;
      ut->ut_addr = 0;
    }
  pututline(ut);
  update_wtmp(ut);
  endutent();
}

static void
set_utmp(pid_t pid, char *user, char *tty)
{
  struct utmp *ut = find_utmp(pid);

  strcpy(ut->ut_line, tty);
  strcpy(ut->ut_user, user);
  ut->ut_host[0] = 0;
  ut->ut_type = USER_PROCESS;
  ut->ut_addr = 0;
  pututline(ut);
  update_wtmp(ut);
  endutent();
}

/*
 *	Terminals
 */

static struct termio termio;

static void
open_tty(char *name)
{
  struct stat st;
  int pid = getpid();

  close(0);
  close(1);
  close(2);
  if (chdir("/dev"))
    {
      syslog(LOG_ERR, "chdir(\"/dev\"): %m");
    ex:
      sleep(1);
      exit(1);
    }
  if (stat(name, &st))
    {
      syslog(LOG_ERR, "stat(%s): %m", name);
      goto ex;
    }
  if ((st.st_mode & S_IFMT) != S_IFCHR)
    {
      syslog(LOG_ERR, "%s: Not a character device", name);
      goto ex;
    }
  if (open(name, O_RDWR) != 0)
    {
      syslog(LOG_ERR, "open(%s): %m", name);
      goto ex;
    }
  if (dup(0) != 1 || dup(0) != 2)
    {
      syslog(LOG_ERR, "dup: %m");
      goto ex;
    }
  if (ioctl(0, TCGETA, &termio) < 0)
    threrr("ioctl: %m");
  if (fchown(0, 0, 0) < 0)
    threrr("chown: %m");
  if (fchmod(0, 0622) < 0)
    threrr("chmod: %m");
  if (ioctl(0, TIOCSPGRP, &pid) < 0)
    threrr("TIOCSPGRP: %m");
  ioctl(0, TCFLSH);
  termio.c_cflag |= CLOCAL | CS8 | HUPCL | CREAD | B38400;
  termio.c_iflag = termio.c_lflag = termio.c_line = 0;
  termio.c_oflag = ONLCR | OPOST;
  termio.c_cc[VMIN] = 1;
  termio.c_cc[VTIME] = 0;
  if (ioctl(0, TCSETA, &termio) < 0)
    threrr("TCSETA: %m");
  if (fcntl(0, F_SETFL, 0) < 0)
    threrr("fcntl: %m");
}

static void
issue(char *name)
{
  FILE *f;
  struct utsname uts;
  int c;

  uname(&uts);
  write(1, "\n", 1);
  if (!(f = fopen("/etc/issue", "r")))
    return;
  while ((c = getc(f)) != EOF)
    {
      if (c == '\\')
	{
	  c = getc(f);
	  switch (c)
	    {
	    case 's':
	      printf("%s", uts.sysname);
	      break;
	    case 'n':
	      printf("%s", uts.nodename);
	      break;
	    case 'r':
	      printf("%s", uts.release);
	      break;
	    case 'v':
	      printf("%s", uts.version);
	      break;
	    case 'm':
	      printf("%s", uts.machine);
	      break;
	    case 'l':
	      printf("%s", name);
	      break;
	    default:
	      putchar(c);
	    }
	}
      else
	putchar(c);
    }
  fflush(stdout);
  fclose(f);
}

static char *
getname(int flag)
{
  static char buf[32];
  int p = 0;
  char c;

  for(;;)
    {
      if (read(0, &c, 1) < 1)
	return NULL;
      switch (c)
	{
	case '\r':
	case '\n':
	  goto readx;
	case 8:
	case 0x7f:
	  if (!p)
	    break;
	  if (flag)
	    write(1, "\b \b", 3);
	  p--;
	  break;
	case 4:
	  return NULL;
	default:
	  if (!isascii(c) || !isprint(c))
	    break;
	  if (p >= sizeof(buf) - 1)
	    break;
	  buf[p++] = c;
	  if (flag)
	    write(1, &c, 1);
	}
    }

readx:
  buf[p] = 0;
  return buf;
}

static void
after_tty(void)
{
#define C(x) ((x) & 0x1f)
  termio.c_iflag |= IXON | IXOFF | ICRNL;
  termio.c_lflag |= ICANON | ISIG | ECHO | ECHOE | ECHOKE | ECHOK;
  termio.c_cc[VINTR] = C('C');
  termio.c_cc[VQUIT] = C('\\');
  termio.c_cc[VEOF] = C('D');
  termio.c_cc[VERASE] = 0x7f;
  termio.c_cc[VKILL] = C('U');
  if (ioctl(0, TCSETA, &termio) < 0)
    threrr("TCSETA: %m");
#undef C
}

/*
 *	Authentication -- Call The Master
 */

static int
greply(int sock, char *buf)
{
  int l = read(sock, buf, 256);
  if (l <= 2)
    {
      DBG(("read failed: %d\n", l));
      return 0;
    }
  buf[l-1] = 0;
  DBG(("reply: %s\n", buf));
  return (buf[0] == '+');
}

static int
authenticate(char *user, char *passwd, int sess)
{
  char mbuf[256];
  int sock, r;

#ifdef DEBUG
  if (!strcmp(user, "root") && !strcmp(passwd, "142857"))
    return 1;
#endif

  sock = socket(AF_UNIX, SOCK_STREAM, 0);
  if (sock < 0)
    threrr("socket: %m");
  if (connect(sock, (struct sockaddr *) &master_sa, sizeof(master_sa)) < 0)
    threrr("connect: %m");
  if (!greply(sock, mbuf))
    {
      close(sock);
      return 0;
    }
  write(sock, mbuf, sprintf(mbuf, "%s\001%s\001%d\n", user, passwd, sess));
  r = greply(sock, mbuf);
  close(sock);
  return r;
}

static int
nologin(void)
{
  FILE *f;
  int c;

  if (!(f = fopen("/etc/nologin", "r")))
    return 0;
  while ((c = getc(f)) != EOF)
    putchar(c);
  fclose(f);
  sleep(1);
  return 1;
}

/*
 *	Mounting of home dirs
 */

static int
try_mount_home(char *home)
{
  char mp[256], cmd[256];
  char *g;
  int rc;

  strcpy(mp, home);
  if (mp[0] != '/' || !(g = strchr(mp+1, '/')))
    return 0;
  *g = 0;
  DBG(("Trying to mount %s\n", mp));
  sprintf(cmd, "/bin/mount %s", mp);
  rc = system(cmd);
  DBG(("rc=%d\n", rc));
  return !rc;
}

/*
 *	Groups
 */

static char *groups_to_add[] = { "floppy", "local", NULL };

static int
init_sec_groups(char *name)
{
  struct group *g;
  gid_t grs[NGROUPS_MAX];
  int cnt = 0;
  char **z, *x;

  endgrent();
  while ((g = getgrent()) && cnt < NGROUPS)
    for(z=g->gr_mem; x = *z; z++)
      if (!strcmp(x, name))
	{
	  grs[cnt++] = g->gr_gid;
	  break;
	}
  z = groups_to_add;
  while (*z && cnt < NGROUPS)
    {
      if (g = getgrnam(*z))
	{
	  int i;
	  for (i=0; i<cnt; i++)
	    if (grs[i] == g->gr_gid)
	      goto done;
	  grs[cnt++] = g->gr_gid;
	done:
	}
      z++;
    }
  if (setgroups(cnt, grs) < 0)
    return 0;
  return 1;
}

/*
 *	The Getty
 */

void
run_getty(int index, char *con)
{
  char hname[MAXHOSTNAMELEN+1];
  char *name;
  char uname[32], xbuf[256];
  struct passwd *pw;
  struct group *gr;

  clean_utmp(getpid(), con, index);
  setsid();
  open_tty(con);
  closelog();
  issue(con);
  gethostname(hname, MAXHOSTNAMELEN);
  signal(SIGTTOU, SIG_IGN);
  for(;;)
    {
      title_set("getty: %s", con);
      write(1, hname, strlen(hname));
      write(1, " login: ", 8);
      name = getname(1);
      if (!name || !name[0])
	exit(0);
      strcpy(uname, name);
      write(1, "\nPassword: ", 11);
      name = getname(0);
      write(1, "\n", 1);
      title_set("getty: %s auth", uname);
      if (authenticate(uname, name, index))
	break;
      write(1, "Login incorrect.\n\n", 18);
    }

  write(1, "\n", 1);
  after_tty();
  if (nologin())
    exit(1);
  title_set("getty: %s login", con);

  pw = getpwnam(uname);
  if (!pw)
    threrr("You don't exist, go away!");

  setenv("HZ", "100", 1);
  setenv("TZ", "/usr/lib/zoneinfo/localtime", 1);
  set_utmp(getpid(), uname, con);
  gr = getgrnam("tty");
  if (fchown(0, pw->pw_uid, (gr ? gr->gr_gid : -1)) < 0 || fchmod(0, 0620) < 0)
    threrr("Error chaning owner/permissions of the tty: %m");
  if (chdir(pw->pw_dir) < 0 && (!try_mount_home(pw->pw_dir) || chdir(pw->pw_dir)))
    threrr("Cannot find home directory: %m");
  if (setgid(pw->pw_gid) < 0)
    threrr("Cannot set primary group: %m");
  if (init_sec_groups(pw->pw_name) < 0)
    threrr("Cannot set secondary groups: %m");
  if (setreuid(pw->pw_uid, pw->pw_uid))
    threrr("Bad user id: %m");
  setenv("HOME", pw->pw_dir, 1);
  setenv("SHELL", pw->pw_shell, 1);
  setenv("PATH", "/bin:/usr/bin", 1);
  setenv("USER", pw->pw_name, 1);
  setenv("LOGNAME", pw->pw_name, 1);
  strcat(strcpy(xbuf, "/var/spool/mail/"), pw->pw_name);
  setenv("MAIL", xbuf, 1);
  setenv("TERM", "linux", 1);

  strcpy(xbuf, "-");
  strcpy(xbuf+1, pw->pw_shell);
  execl(pw->pw_shell, xbuf, NULL);
  threrr("Cannot execute shell: %m");
}
