/*
 * This module contains support for the Simple User Authentication System
 *
 * Copyright (c) 1996--1999 Martin Mares  <mj@atrey.karlin.mff.cuni.cz>
 *
 * This software maybe be used for any purpose provided
 * the above copyright notice is retained.  It is supplied
 * as is, with no warranty expressed or implied.
 */

#include <pwd.h>
#include <grp.h>
#include <time.h>

#include "nfsd.h"

#define UID_HASH_SIZE 64
#define UID_HASH(x) ((x) & (UID_HASH_SIZE - 1))
#define UID_MAX_GRP 16
#define UID_REFRESH_TIME 30

struct uidinfo {
	struct uidinfo *next;
	uid_t uid;
	short ngrp;
	time_t rfsh;
	gid_t grp[UID_MAX_GRP];
};

static struct uidinfo *uidhash[UID_HASH_SIZE];

struct userinfo {
	struct userinfo *next;
	struct uidinfo *uid;
};

struct ipinfo {
	struct ipinfo *next;
	struct userinfo *users;
	unsigned ip;
};

static struct ipinfo *first_ip;

static inline void
add_gid(u, gid)
struct uidinfo *u;
gid_t gid;
{
	Dprintf(D_AUTH, "Adding gid %d for uid %d\n", gid, u->uid);
	if (u->ngrp < UID_MAX_GRP)
		u->grp[u->ngrp++] = gid;
}

static void
refresh_uidinfo(u)
struct uidinfo *u;
{
	struct passwd *p;
	struct group *g;
	char **gg;

	Dprintf(D_AUTH, "Refreshing uid %d\n", u->uid);
	u->ngrp = 0;
	setpwent();
	p = getpwuid(u->uid);
	if (p) {
		add_gid(u, p->pw_gid);
		setgrent();
		while ((g = getgrent())) {
			gg = g->gr_mem;
			while (*gg) {
				if (!strcmp(*gg, p->pw_name)) {
					add_gid(u, g->gr_gid);
					break;
				}
				gg++;
			}
		}
		endgrent();
	}
	endpwent();
}

static struct uidinfo *
get_uidinfo(uid)
uid_t uid;
{
	unsigned i = UID_HASH(uid);
	struct uidinfo *u;
	time_t now = time(NULL);

	u = uidhash[i];
	while (u) {
		if (u->uid == uid)
			break;
		u = u->next;
	}
	if (!u) {
		u = xmalloc(sizeof(struct uidinfo));
		u->next = uidhash[i];
		uidhash[i] = u;
		u->uid = uid;
	} else if (u->rfsh + UID_REFRESH_TIME >= now)
		return u;

	u->rfsh = now;
	refresh_uidinfo(u);
	return u;
}

static struct ipinfo *
get_ip(ip)
unsigned ip;
{
	struct ipinfo *i = first_ip;

	while (i) {
		if (i->ip == ip)
			return i;
		i = i->next;
	}
	return NULL;
}

static struct ipinfo *
new_ip(ip)
unsigned ip;
{
	struct ipinfo *i;

	if ((i = get_ip(ip)))
		return i;
	i = xmalloc(sizeof(struct ipinfo));
	i->next = first_ip;
	first_ip = i;
	i->users = NULL;
	i->ip = ip;
	return i;
}

static void
suas_add_uid(ip, uid)
unsigned	ip;
uid_t		uid;
{
	struct ipinfo *i;
	struct userinfo *u;
	struct uidinfo *d;

	Dprintf(D_AUTH, "Adding user %d for IP %08lx\n", uid, ip);

	i = new_ip(ip);
	if (!i)
		return;

	for(u = i->users; u; u = u->next)
		if (u->uid->uid == uid)
			return;

	d = get_uidinfo(uid);
	if (!d)
		return;

	u = xmalloc(sizeof(struct uidinfo));
	u->next = i->users;
	i->users = u;
	u->uid = d;
}

static void
suas_rem_uid(ip, uid)
unsigned	ip;
uid_t		uid;
{
	struct ipinfo *i;
	struct userinfo *u, **v;

	Dprintf(D_AUTH, "Removing user %d from IP %08lx\n", uid, ip);

	i = get_ip(ip);
	if (!i)
		return;
	v = &i->users;
	while ((u = *v)) {
		if (u->uid->uid == uid) {
			*v = u->next;
			break;
		}
		v = &u->next;
	}
}

static void
suas_remove_ip(struct ipinfo *i)
{
	struct userinfo *u;

	Dprintf(D_AUTH, "Removing IP %08x\n", i->ip);

	while ((u = i->users)) {
		i->users = u->next;
		free(u);
	}
}

static void
suas_rem_ip(ip)
unsigned	ip;
{
	struct ipinfo *i;

	Dprintf(D_AUTH, "Requested remove IP %08x\n", ip);

	if (ip) {
		i = get_ip(ip);
		if (i)
			suas_remove_ip(i);
	} else {
		while ((i = first_ip)) {
			first_ip = i->next;
			suas_remove_ip(i);
			free(i);
		}
	}
}

void
suas_set(ip, op, uid)
unsigned	ip;
uicop		op;
uid_t		uid;
{
	switch (op) {
		case UICADD:
			suas_add_uid(ip, uid);
			break;
		case UICREM:
			suas_rem_uid(ip, uid);
			break;
		case UICCLEAR:
			suas_rem_ip(ip);
			break;
		default:
			Dprintf(L_ERROR, "Invalid SUAS UID mapping requested (%d)\n", op);
	}
}

static struct uidinfo *
lookup_uid(ip, uid)
unsigned	ip;
uid_t		uid;
{
	struct ipinfo *i;
	struct userinfo *u;

	i = get_ip(ip);
	if (!i)
		return NULL;
	u = i->users;
	while (u) {
		if (u->uid->uid == uid)
			return u->uid;
		u = u->next;
	}
	return NULL;
}

static int
lookup_gid(u, gid)
struct uidinfo * u;
gid_t		gid;
{
	int i;

	if (!u)
		return 0;
	for(i=0; i<u->ngrp; i++)
		if (u->grp[i] == gid)
			return 1;
	return 0;
}

uid_t
suas_map_uid(uid, mountp, ip)
uid_t		uid;
nfs_mount *	mountp;
unsigned	ip;
{
	uid_t nuid;

	if (uid == mountp->o.nobody_uid || !lookup_uid(ip, uid))
		nuid = mountp->o.nobody_uid;
	else
		nuid = uid;

	Dprintf(D_AUTH, "suas_map_uid(%08x,%d) -> %d\n", ip, uid, nuid);
	return nuid;
}

gid_t
suas_map_gid(gid, mountp, ip, uid)
gid_t		gid;
nfs_mount *	mountp;
unsigned	ip;
uid_t		uid;
{
	gid_t ngid;
	struct uidinfo *u;

	if (uid == mountp->o.nobody_uid || gid == mountp->o.nobody_gid)
		ngid = mountp->o.nobody_gid;
	else {
		u = lookup_uid(ip, uid);
		if (lookup_gid(u, gid))
			ngid = gid;
		else
			ngid = mountp->o.nobody_gid;
	}

	Dprintf(D_AUTH, "suas_map_gid(%08x,%d,%d) -> %d\n", ip, gid, uid, ngid);
	return ngid;
}
