/*
 *	Simple User Authentication System -- Client
 *
 *	(c) 1996--1998 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License. See file COPYING in any of the GNU packages.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <pwd.h>
#include <alloca.h>

#include "suas.h"

static char *server_name;

char program_name[] = "suac";

static char *options = "s:a:d:cvwSC:t";

static char *help_string = "\
Simple User Authentication System Client " SUA_VERSION " (c) " SUA_YEAR " Martin Mares\n\
\n\
Usage: suac [-s <server>] [-v] <command>\n\
\n\
-a <user>\tCreate new user\n\
-d <user>\tDelete a user\n\
-c\t\tChange password\n\
-C <user>\tChange password of specified user\n\
-t\t\tTest password\n\
-w\t\tList logged in users\n\
-v\t\tBe verbose\n\
-S\t\tUse gets(stdin) instead of getpass()\n\
\n";

static int gets_flag;

static void
help(void)
{
  fprintf(stderr, help_string);
  exit(1);
}

static int sock;
static FILE *fi, *fo;
static int verbose;
static char inbuf[256];
static char *sesskey;

static char *
reply(void)
{
  char *c, *d, *e;

  fflush(fo);
  if (!fgets(inbuf, 256, fi))
    die("Connection to server lost");
  if (c = strchr(inbuf, '\n'))
    *c = 0;
  if (c = strchr(inbuf, '\r'))
    *c = 0;
  if (verbose)
    fprintf(stderr, "Got reply %s\n", inbuf);
  if (inbuf[0] != '+')
    die("Server error: %s", inbuf+1);
  c = strchr(inbuf, '<');
  if (!c)
    return NULL;
  d = strrchr(inbuf, '>');
  if (!d)
    return NULL;
  *c++ = 0;
  *d = 0;
  e = xmalloc(strlen(c) + 1);
  strcpy(e, c);
  return e;
}

static char *
getpassw(char *a)
{
  char *m, *z;

  if (gets_flag)
    {
      z = alloca(256);
      if (!fgets(z, 256, stdin))
	die("Password read error");
      if (!(m = strchr(z, '\n')))
	die("No password");
      *m = 0;
    }
  else
    z = getpass(a);
  m = xmalloc(1 + strlen(z));
  strcpy(m, z);
  return m;
}

static char *
getpwd(void)
{
  char *a = getpassw("New password: ");
  char *b = getpassw("Again, please: ");
  char *c;

  if (strcmp(a, b))
    die("Passwords don't match");
  c = xmalloc(1 + SUA_USER_KEY_SIZE);
  encrypt_first(a, c, "");
  return c;
}

static void
cmd_add(char *name)
{
  char *p = getpwd();
  fprintf(fo, "adduser %s %s\n", name, p);
  reply();
}

static void
cmd_del(char *name)
{
  fprintf(fo, "deluser %s\n", name);
  reply();
}

static char *
whoami(void)
{
  struct passwd *h;

  if (h = getpwuid(getuid()))
    return h->pw_name;
  die("You don't exist, go away");
}

static void
cmd_passwd(char *user)
{
  char *w = user ? : whoami();
  char *salt, *pwd, *nep;
  byte uk[SUA_USER_KEY_SIZE+1];
  byte nk[SUA_NET_KEY_SIZE+1];
  byte rk[SUA_USER_KEY_SIZE+1];

  fprintf(fo, "salt %s\n", w);
  salt = reply();
  if (!user)
    {
      pwd = getpassw("Old password: ");
      encrypt_first(pwd, rk, salt);
      encrypt_local(rk, uk, salt);
      encrypt_net(uk, nk, sesskey);
    }
  else
    strcpy(nk, "?");
  nep = getpwd();
  fprintf(fo, "passwd %s %s %s\n", w, nk, nep);
  reply();
}

static void
cmd_who(void)
{
  struct sockaddr_in si;
  char *c, *d, *e;
  struct hostent *he;
  char buf[16];

  fprintf(fo, "who\n");
  for(;;)
    {
      reply();
      c = inbuf;
      while (*c && *c != '\t')
	c++;
      if (!*c)
	break;
      *c++ = 0;
      e = c;
      while (*e && *e != '\t')
	e++;
      if (!*e)
	break;
      *e++ = 0;
      sscanf(inbuf+1, "%x", &si.sin_addr.s_addr);
      he = gethostbyaddr((char *) &si.sin_addr.s_addr, sizeof(si.sin_addr.s_addr), AF_INET);
      if (he)
	d = he->h_name;
      else
	{
	  unsigned l = si.sin_addr.s_addr;
	  sprintf(buf, "%d.%d.%d.%d",
		  (l >> 24) & 255,
		  (l >> 16) & 255,
		  (l >> 8) & 255,
		  l & 255);
	  d = buf;
	}
      printf("%-20s%3s %s\n", d, e, c);
    }
}

static void
cmd_test(void)
{
  char *w = whoami();
  char *salt, *pwd;
  byte uk[SUA_USER_KEY_SIZE+1];
  byte nk[SUA_NET_KEY_SIZE+1];
  byte rk[SUA_USER_KEY_SIZE+1];

  fprintf(fo, "salt %s\n", w);
  salt = reply();
  pwd = getpassw("Password: ");
  encrypt_first(pwd, rk, salt);
  encrypt_local(rk, uk, salt);
  encrypt_net(uk, nk, sesskey);
  fprintf(fo, "auth %s %s\n", w, nk);
  reply();
}

int
main(int argc, char **argv)
{
  int i;
  char *user = NULL;
  int op = 0;
  int need_local = 0;
  int need_root = 0;
  int we_are_root;

  while ((i = getopt(argc, argv, options)) != -1)
    switch (i)
      {
      case 's':
	server_name = optarg;
	break;
      case 'a':
	user = optarg;
	op = 1;
	need_local = need_root = 1;
	break;
      case 'd':
	user = optarg;
	op = 2;
	need_local = need_root = 1;
	break;
      case 'c':
	op = 3;
	need_local = 1;
	break;
      case 'C':
	user = optarg;
	op = 3;
	need_root = need_local = 1;
	break;
      case 'w':
	op = 4;
	break;
      case 'v':
	verbose++;
	break;
      case 'S':
	gets_flag = 1;
	break;
      case 't':
	op = 5;
	break;
      default:
	help();
      }

  if (optind < argc)
    help();
  if (!op)
    die("Nothing happens");

  we_are_root = !getuid();
  if (we_are_root < need_root)
    die("You are not allowed to do this");
  if (need_local && server_name)
    die("Need local connection to do this");
  if (need_local && need_root)
    {
      struct sockaddr_un addr;

      sock = socket(AF_UNIX, SOCK_STREAM, 0);
      if (sock < 0)
	die("socket(): %m");
      addr.sun_family = AF_UNIX;
      strcpy(addr.sun_path, SUA_DAEMON_SOCKET);
      if (connect(sock, (struct sockaddr *) &addr, sizeof(addr)) < 0)
	die("Cannot connect to local server: %m");
    }
  else
    {
      struct hostent *he;
      struct sockaddr_in addr;

      if (need_local || !(server_name = getenv("SUAS_SERVER")))
	server_name = "localhost";
      sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
      if (sock < 0)
	die("socket(): %m");
      if (!(he = gethostbyname(server_name)))
	die("Cannot resolve host name `%s': %m", server_name);
      addr.sin_family = AF_INET;
      addr.sin_port = htons(SUA_SERVER_PORT);
      memcpy(&addr.sin_addr.s_addr, he->h_addr, he->h_length);
      if (connect(sock, (struct sockaddr *) &addr, sizeof(addr)) < 0)
	die("Cannot connect to `%s': %m", server_name);
    }

  fi = fdopen(sock, "r");
  fo = fdopen(sock, "w");
  if (!fi || !fo)
    die("fdopen failed");

  sesskey = reply();

  switch (op)
    {
    case 1:
      cmd_add(user);
      break;
    case 2:
      cmd_del(user);
      break;
    case 3:
      cmd_passwd(user);
      break;
    case 4:
      cmd_who();
      break;
    case 5:
      cmd_test();
      break;
    }

  return 0;
}
