/*
 *	Simple User Authentication System -- Server
 *
 *	(c) 1996--1998 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License. See file COPYING in any of the GNU packages.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>
#include <time.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <rpc/types.h>
#include <rpc/xdr.h>

#include "suas.h"
#ifdef DEBUG
#define DBG(x) printf x
#define DBG2(x)
#else
#define DBG(x)
#define DBG2(x)
#endif

#ifndef IN_LOOPBACK			/* FIXME: Do we need this? */
#define IN_LOOPBACK(a) (((a) & 0x7f000000) == 0x7f000000)
#endif

char program_name[] = "suad";

/*
 *	Communication with NFS daemon.
 */

#ifdef NFS_HACK

#include "nfsd/nfs_prot.h"

/* Fake XDR functions */

bool_t
xdr_nfsstat(XDR *xdrs, nfsstat *objp)
{
  if (!xdr_enum(xdrs, (enum_t *)objp))
    return FALSE;
  return TRUE;
}

bool_t
xdr_uidctrlargs(XDR *xdrs, uidctrlargs *objp)
{
  if (!xdr_u_int(xdrs, &objp->ipaddr))
    return FALSE;
  if (!xdr_enum(xdrs, (enum_t *)&objp->op))
    return FALSE;
  if (!xdr_u_int(xdrs, &objp->uid))
    return FALSE;
  return TRUE;
}

/* NFS interface */

#define NFS_PORT 2049

static CLIENT *mclient;
static struct timeval total_timeout, pertry_timeout;

static void
init_nfs(void)
{
  struct sockaddr_in sa, sb;
  int msock;
  int one = 1;

  total_timeout.tv_sec = 30;
  total_timeout.tv_usec = 0;
  pertry_timeout.tv_sec = 2;
  pertry_timeout.tv_usec = 0;

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
  sa.sin_port = htons(NFS_PORT);
  sb.sin_family = AF_INET;
  sb.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
  sb.sin_port = htons(SUA_SERVER_PORT);

  msock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (msock < 0)
    {
      syslog(LOG_ERR, "socket: %m");
      exit(1);
    }
  if (setsockopt(msock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) < 0)
    {
      syslog(LOG_ERR, "setsockopt: %m");
      exit(1);
    }
  if (bind(msock, (struct sockaddr *) &sb, sizeof(sb)))
    {
      syslog(LOG_ERR, "bind: %m");
      exit(1);
    }

  if (!(mclient = clntudp_create(&sa, NFS_PROGRAM, NFS_VERSION, pertry_timeout, &msock)))
    {
      syslog(LOG_ERR, "Unable to connect to the NFS server");
      exit(1);
    }
  mclient->cl_auth = authnone_create();
}

static void
cmd_nfs(unsigned ip, uicop op, unsigned uid)
{
  struct uidctrlargs ua;
  int res;
  enum nfsstat stat;

  ua.ipaddr = ip;
  ua.op = op;
  ua.uid = uid;
  res = clnt_call(mclient, NFSPROC_UID_CONTROL,
		  (xdrproc_t) xdr_uidctrlargs, &ua,
		  (xdrproc_t) xdr_nfsstat, &stat,
		  total_timeout);
  if (res != RPC_SUCCESS)
    {
      syslog(LOG_ERR, "RPC call failed: %d", res);
      return;
    }
  if (stat != NFS_OK)
    {
      syslog(LOG_ERR, "NFS_UID_CTRL failed: %d", stat);
      return;
    }
}

#else

#define init_nfs()
#define cmd_nfs(x,y,z)

#endif

/* User list */

struct user_info {
  struct user_info *next;
  char *name, *salt, *passwd;
  ulg uid;
};

static struct user_info *first_user_info;

static char *
alloc_string(char *s)
{
  char *b = xmalloc(strlen(s) + 1);
  strcpy(b, s);
  return b;
}

static int
scan_passwd(struct user_info *ui, int warnflag)
{
  FILE *f;
  char buf[1024];
  ulg uid;
  int flag;
  struct user_info *u;

  f = fopen(SUA_PASSWD, "r");
  if (!f)
    {
      syslog(LOG_ERR, "Cannot open " SUA_PASSWD ": %m");
      exit(1);
    }
  while (fgets(buf, sizeof(buf), f))
    {
      char *s = strchr(buf, '\n');
      if (!s)
	{
	fault:
	  syslog(LOG_ERR, SUA_PASSWD ": Line too long or malformed");
	  break;
	}
      *s = 0;
      s = buf;
      while (*s && *s != ':')
	s++;
      if (!*s)
	goto fault;
      *s++ = 0;
      while (*s && *s != ':')
	s++;
      if (!*s++)
	goto fault;
      sscanf(s, "%d", &uid);
      for(u=ui; u; u=u->next)
	if (u->uid == ~0 && !strcmp(buf, u->name))
	  {
	    u->uid = uid;
	    break;
	  }
    }
  fclose(f);

  flag = 1;
  for(u=ui; u; u=u->next)
    if (u->uid == ~0)
      {
	flag = 0;
	if (warnflag)
	  syslog(LOG_ERR, "User `%s' not found in passwd file", u->name);
      }
  return flag;
}

static void
read_users(void)
{
  struct user_info *u;
  char buf[256];
  FILE *f;
  char *z, *a, *b;
  int lino = 0;

  f = fopen(SUA_USER_LIST_FILE, "r");
  if (f)
    {
      for(;;)
	{
	  lino++;
	  fgets(buf, 254, f);
	  if (feof(f))
	    break;
	  z = strchr(buf, '\n');
	  if (!z)
	    {
	    malf:
	      syslog(LOG_ERR, "Line %d of user list file malformed", lino);
	      continue;
	    }
	  *z = 0;
	  if (!buf[0] || buf[0] == '#')
	    continue;
	  a = strchr(buf, ':');
	  if (!a)
	    goto malf;
	  *a++ = 0;
	  b = strchr(a, ':');
	  if (!b)
	    goto malf;
	  *b++ = 0;
	  if (strlen(a) > SUA_USER_SALT_SIZE ||
	      strlen(b) > SUA_USER_KEY_SIZE)
	    goto malf;
	  u = xmalloc(sizeof(struct user_info));
	  u->name = alloc_string(buf);
	  u->salt = alloc_string(a);
	  u->passwd = alloc_string(b);
	  u->uid = ~0;
	  u->next = first_user_info;
	  first_user_info = u;
	}
      fclose(f);
    }
  else
    syslog(LOG_WARNING, "Unable to open %s, using null auth file", SUA_USER_LIST_FILE);
  scan_passwd(first_user_info, 1);
}

static void
write_users(void)
{
  FILE *f;

  rename(SUA_USER_LIST_FILE, SUA_USER_LIST_BAK_FILE);
  f = fopen(SUA_USER_LIST_FILE, "w");
  fchmod(fileno(f), 0600);
  if (f)
    {
      struct user_info *u = first_user_info;
      struct tm *tm;
      time_t now = time(NULL);

      tm = localtime(&now);
      fprintf(f, "# Dumped automatically on %02d/%02d/%04d, %02d:%02d:%02d\n",
	      tm->tm_mday, tm->tm_mon+1, tm->tm_year+1900,
	      tm->tm_hour, tm->tm_min, tm->tm_sec);
      while (u)
	{
	  fprintf(f, "%s:%s:%s\n", u->name, u->salt, u->passwd);
	  u = u->next;
	}
      fclose(f);
    }
  else
    syslog(LOG_ERR, "Cannot write %s", SUA_USER_LIST_FILE);
}

struct user_info *
find_user(char *name)
{
  struct user_info *u = first_user_info;

  while (u)
    {
      if (!strcmp(u->name, name) && u->uid != ~0)
	return u;
      u = u->next;
    }
  return NULL;
}

struct user_info *
find_user_uid(ulg uid)
{
  struct user_info *u = first_user_info;

  while (u)
    {
      if (u->uid == uid)
	return u;
      u = u->next;
    }
  return NULL;
}

/* Domains */

struct domrec {
  struct domrec *next;
  ulg addr, mask;
};

static struct domrec *first_client;
static struct domrec **last_client = &first_client;

static int
parse_domain(struct domrec *a, char *buf)
{
  ulg i,j;
  char *k;
  int z;

  i = 0;
  for(j=0; j<4; j++)
    {
      if (!*buf)
	return 0;
      k = buf;
      while (*buf && *buf != '.' && *buf != '/')
	buf++;
      if (k == buf)
	return 0;
      z = 0;
      while (k < buf)
	{
	  if (*k < '0' || *k > '9')
	    return 0;
	  z = z*10 + *k - '0';
	  if (z > 255)
	    return 0;
	  k++;
	}
      i = (i << 8) | z;
      if (*buf == '.')
	buf++;
    }

  a->addr = htonl(i);
  if (*buf == '/')
    {
      buf++;
      z = 0;
      while (*buf)
	{
	  if (*buf < '0' || *buf > '9')
	    return 0;
	  z = z*10 + *buf++ - '0';
	  if (z > 32)
	    return 0;
	}
      a->mask = htonl(~((1 << (32 - z)) - 1));
    }
  else
    a->mask = ~0;

  DBG(("Adding %08x/%08x\n", a->addr, a->mask));

  return 1;
}

static void
read_domains(void)
{
  FILE *f;
  char buf[256];
  char *c;
  struct domrec *a;
  int lin = 0;

  if (!(f = fopen(SUA_CLIENT_LIST_FILE, "r")))
    {
      syslog(LOG_ERR, "Cannot read %s", SUA_CLIENT_LIST_FILE);
      exit(1);
    }
  while (fgets(buf, 256, f))
    {
      lin++;
      c = strchr(buf, '\n');
      if (!*c)
	syslog(LOG_ERR, "Line too long in %s", SUA_CLIENT_LIST_FILE);
      *c = 0;
      if (!buf[0] || buf[0] == '#')
	continue;
      a = xmalloc(sizeof(struct domrec));
      *last_client = a;
      a->next = NULL;
      last_client = &a->next;
      if (!parse_domain(a, buf))
	{
	  syslog(LOG_ERR, "Line %d in " SUA_CLIENT_LIST_FILE " malformed", lin);
	  exit(1);
	}
    }
  fclose(f);
}

static int
match_client(ulg addr)
{
  struct domrec *a = first_client;

  while (a)
    {
      if (!((a->addr ^ addr) & a->mask))
	return 1;
      a = a->next;
    }
  return 0;
}

/* Connections */

static int sock;

struct connection {
  struct connection *next, *prev;
  struct in_addr client;
  int sock;
  int islocal, local_uid;
  int state, waiting, timeout, id, ismaster;
  ulg port, tmp;
  byte rbuf[256], wbuf[256], sesskey[SUA_SESSION_KEY_SIZE+1];
  byte *rbptr, *rbend, *wbptr, *wbend;
};

#define STATE_ACCEPT 0
#define STATE_HELLO 1
#define STATE_COMMAND 2
#define STATE_DROP 3
#define STATE_WAIT 4
#define STATE_WHO 5
#define STATE_MAX 6

#define WAIT_READ 1
#define WAIT_WRITE 2
#define WAIT_TIMER 4

static struct connection *first_connection;

static int num_connects, conn_id;

#define MAX_CONNECTIONS 64

static char *
conn_name(struct connection *c)
{
  static char buf[16];
  byte *k = (byte *) &c->client.s_addr;

  sprintf(buf, "%d.%d.%d.%d", k[0], k[1], k[2], k[3]);
  return buf;
}

static inline void
timeout(struct connection *c, ulg tim)
{
  c->timeout = time(NULL) + tim;
}

static char *
get_string(struct connection *c)
{
  byte *p, *q;

  if (c->rbptr != c->rbuf)
    {
      int s = c->rbptr - c->rbuf;
      memmove(c->rbuf, c->rbptr, s);
      c->rbptr -= s;
      c->rbend -= s;
    }

  p = q = c->rbptr;
  while (p <= c->rbend)
    {
      if (*p == '\n')
	{
	  if (p > c->rbuf && p[-1] == '\r')
	    p[-1] = 0;
	  *p++ = 0;
	  c->rbptr = p;
	  return q;
	}
      p++;
    }
  c->waiting = WAIT_READ;
  timeout(c, SUA_TIMEOUT);
  return NULL;
}

static void
waiting(struct connection *c, ulg tim)
{
  timeout(c, tim);
  c->waiting = WAIT_TIMER;
}

static void
put_string(struct connection *c, char *s)
{
  if (c->wbend + strlen(s) >= c->wbuf + sizeof(c->wbuf))
    {
      syslog(LOG_CRIT, "Output buffer overflow");
      exit(1);
    }
  while (*s)
    *c->wbend++ = *s++;
  c->waiting = WAIT_WRITE;
  timeout(c, SUA_TIMEOUT);
}

#define GO(s) c->state = STATE_##s;

/* Logging */

static FILE *log;

static void
log_open(void)
{
  log = fopen(SUA_LOG_FILE, "a+");
  if (!log)
    {
      syslog(LOG_CRIT, "Unable to open log file");
      exit(1);
    }
}

#define LOG_REPEAT_TIMEOUT 16

static void
do_log(char *msg, ...)
{
  va_list args;
  struct tm *tm;
  time_t now;
  int same;
  byte timebuf[32], buf[1024];
  static byte lastmsg[1024];
  static int lastmsg_count;
  static time_t lastmsg_time;

  va_start(args, msg);
  vsprintf(buf, msg, args);
  now = time(NULL);
  same = !strcmp(lastmsg, msg);
  tm = localtime(&now);
  strftime(timebuf, sizeof(timebuf), "%b %d %H:%M:%S ", tm);
  if (same || lastmsg_count && (now - lastmsg_time) >= LOG_REPEAT_TIMEOUT)
    {
      if (!lastmsg_count)
	lastmsg_time = now;
      lastmsg_count++;
      fprintf(log, "%sLast message repeated %d times\n", timebuf, lastmsg_count);
      lastmsg_count = 0;
      if (same)
	goto end;
    }
  strcpy(lastmsg, buf);
  lastmsg_count = 0;
  fprintf(log, "%s%s\n", timebuf, buf);
end:
  fflush(log);
  va_end(args);
}

/* Sessions */

struct logged_in {
  struct logged_in *next;
  int id;
};

struct uid_rec {
  struct uid_rec *next, **prev;
  ulg uid, addr;
  struct logged_in *in;
};

static struct uid_rec *first_uid_rec;

static void
drop_user(struct uid_rec *w)
{
  struct logged_in *m;

  DBG(("Dropping %d from %08x\n", w->uid, w->addr));

  *w->prev = w->next;
  if (w->next)
    w->next->prev = w->prev;
  while (m = w->in)
    {
      w->in = m->next;
      free(m);
    }
  free(w);

  cmd_nfs(w->addr, UICREM, w->uid);
}

static void
add_user(struct uid_rec *w)
{
  DBG(("Adding %d to %08x\n", w->uid, w->addr));

  cmd_nfs(w->addr, UICADD, w->uid);
}

static void
dropping_master(struct connection *c)
{
  struct uid_rec *z, *w;

  DBG(("Dropping master %d for %s\n", c->id, conn_name(c)));
  for(w = first_uid_rec; w; w = z)
    {
      z = w->next;
      if (w->addr == c->client.s_addr)
	drop_user(w);
    }
}

static void
register_user(ulg uid, ulg mach, int id)
{
  struct uid_rec *a;
  struct logged_in *l;

  for(a = first_uid_rec; a; a = a->next)
    if (a->addr == mach && a->uid == uid)
      goto ihave;
  a = xmalloc(sizeof(struct uid_rec));
  a->uid = uid;
  a->addr = mach;
  if (first_uid_rec)
    first_uid_rec->prev = &a->next;
  a->next = first_uid_rec;
  first_uid_rec = a;
  a->prev = &first_uid_rec;
  a->in = NULL;
  add_user(a);
ihave:
  for(l = a->in; l; l=l->next)
    if (l->id == id)
      return;
  l = xmalloc(sizeof(struct logged_in));
  l->next = a->in;
  a->in = l;
  l->id = id;
}

static void
unregister_user(ulg mach, int id)
{
  struct uid_rec *a;
  struct logged_in *l, **m;

  for(a = first_uid_rec; a; a=a->next)
    if (a->addr == mach)
      {
	for(m = &a->in; l = *m; m=&l->next)
	  if (l->id == id)
	    goto delhim;
      }
  return;

delhim:
  *m = l->next;
  if (!a->in)
    drop_user(a);
}

static struct connection *
find_master(struct connection *c)
{
  struct connection *d;

  for(d = first_connection; d; d = d->next)
    if (d->client.s_addr == c->client.s_addr && d->ismaster)
      break;
  return d;
}

/* Commands */

#define MAXARGS 16

static char *arg[MAXARGS];

struct command {
  char *name;
  int nargs, auth;
  int (*cmd)(struct connection *c);
};

#define AU_LOCAL 1
#define AU_PORT 2
#define AU_ROOT 4

static void
dropping(struct connection *c, char *txt)
{
  put_string(c, txt);
  GO(DROP);
}

static int
cmd_quit(struct connection *c)
{
  dropping(c, "+Good bye\n");
  return 1;
}

static void
modify_passwd(struct user_info *u, char *pwd)
{
  generate_random(u->salt, SUA_USER_BIN_SALT_SIZE);
  encrypt_local(pwd, u->passwd, u->salt);
}

static int
check_pass(struct user_info *u, char *passwd, char *sesskey)
{
  byte buf[SUA_NET_KEY_SIZE+1];
  char *p = u->passwd;

  if (p[0] == '$')
    {
#ifdef USE_LOCAL_PASSWORDS
      if (!strcmp(p, "$local"))
	{
	  struct spwd *sp;
	  endspent();
	  if (!(sp = getspnam(u->name)))
	    return 0;
	  encrypt_net(sp->sp_pwdp, buf, sesskey);
	}
      else
#endif
	return 0;
    }
  else
    encrypt_net(u->passwd, buf, sesskey);
  return !strcmp(buf, passwd);
}

static int
cmd_add(struct connection *c)
{
  struct user_info *u, ut;

  if (find_user(arg[1]))
    {
      put_string(c, "-User already known\n");
      return 1;
    }

  ut.next = NULL;
  ut.name = arg[1];
  ut.uid = ~0;
  if (!scan_passwd(&ut, 0))
    {
      put_string(c, "-No such user ID\n");
      return 1;
    }

  do_log("Adding user %s", arg[1]);

  u = xmalloc(sizeof(struct user_info));
  u->name = alloc_string(arg[1]);
  u->uid = ut.uid;
  u->next = first_user_info;
  first_user_info = u;
  u->salt = xmalloc(SUA_USER_SALT_SIZE + 1);
  u->passwd = xmalloc(SUA_USER_KEY_SIZE + 1);
  modify_passwd(u, arg[2]);
  write_users();
  return 0;
}

static int
cmd_del(struct connection *c)
{
  struct user_info *u, *v;

  u = find_user(arg[1]);
  if (!u)
    {
      put_string(c, "-No such user\n");
      return 1;
    }

  do_log("Deleting user %s", arg[1]);
  if (first_user_info == u)
    first_user_info = u->next;
  else
    for(v=first_user_info; v; v=v->next)
      if (v->next == u)
	{
	  v->next = u->next;
	  break;
	}
  /*
   * Note that we don't free the structure here. This really
   * is a memory leak, but it doesn't do any harm. Normal users
   * cannot misuse it to fill system memory as only root is
   * allowed to delete users.
   */
  write_users();
  return 0;
}

static int
cmd_salt(struct connection *c)
{
  struct user_info *u = find_user(arg[1]);

  put_string(c, "+Salt: <");
  if (u)
    {
      if (u->passwd[0] == '$')
	{
	  char salt[SUA_USER_SALT_SIZE + 1];
	  strcpy(salt, "?");
#ifdef USE_LOCAL_PASSWORDS
	  if (!strcmp(u->passwd, "$local"))
	    {
	      struct spwd *sp;
	      endspent();
	      if (sp = getspnam(u->name))
		{
		  char *pw = sp->sp_pwdp;
		  char *g;
		  if (strncmp(pw, "$1$", 3) || !(g = strchr(pw+3, '$')))
		    g = pw+2;
		  else
		    g++;
		  strcpy(salt, "$L");
		  strncpy(salt+2, pw, g-pw);
		  salt[g-pw+2] = 0;
		}
	    }
#endif
	  put_string(c, salt);
	}
      else
	put_string(c, u->salt);
    }
  else
    {
      byte false_clue[SUA_USER_KEY_SIZE + 1];
      encrypt_local(arg[1], false_clue, "Pi!");
      put_string(c, false_clue);
    }
  put_string(c, ">\n");
  return 1;
}

static int
cmd_auth(struct connection *c)
{
  struct user_info *u = find_user(arg[1]);

  if (u && check_pass(u, arg[2], c->sesskey))
    {
      do_log("AUTH %s from %s OK", arg[1], conn_name(c));
      return 0;
    }

  do_log("AUTH %s from %s failed", arg[1], conn_name(c));
  syslog(LOG_WARNING, "AUTH %s from %s failed\n", arg[1], conn_name(c));
  waiting(c, SUA_RETRY_TIMEOUT);
  GO(WAIT);
  return 1;
}

static int
cmd_login(struct connection *c)
{
  struct user_info *u = find_user(arg[1]);
  struct connection *m;
  int tty = atol(arg[3]);

  m = find_master(c);
  if (!m)
    {
      put_string(c, "-Master connection not running\n");
      return 1;
    }
  if (strcmp(m->sesskey, arg[4]))
    {
      put_string(c, "-Invalid master key\n");
      return 1;
    }

  if (u && check_pass(u, arg[2], c->sesskey))
    {
      do_log("LOGIN %s from %s(%d) OK", arg[1], conn_name(c), tty);
      register_user(u->uid, c->client.s_addr, tty);
      return 0;
    }

  do_log("LOGIN %s from %s(%d) failed", arg[1], conn_name(c), tty);
  syslog(LOG_WARNING, "LOGIN %s from %s(%d) failed\n", arg[1], conn_name(c), tty);
  waiting(c, SUA_RETRY_TIMEOUT);
  GO(WAIT);
  return 1;
}

static int
cmd_clear(struct connection *c)
{
  int tty = atol(arg[1]);

  if (!c->ismaster)
    {
      put_string(c, "-Master connection required\n");
      return 1;
    }
  unregister_user(c->client.s_addr, tty);
  do_log("CLEAR %d on %s", tty, conn_name(c));
  return 0;
}

static void drop_conn(struct connection *c, int why);

static int
cmd_master(struct connection *c)
{
  struct connection *m;

  DBG(("%d: Master connection for %s\n", c->id, conn_name(c)));
  do_log("Connected master for %s", conn_name(c));
  if (m = find_master(c))
    {
      /* Need to drop the old master connection */
      DBG(("Dropping old master %d\n", m->id));
      m->state = STATE_DROP;
      m->waiting = 0;
      m->ismaster = 0;
    }

  c->ismaster = 1;
  return 0;
}

static int
cmd_passwd(struct connection *c)
{
  struct user_info *u = find_user(arg[1]);

  if (u && (!c->local_uid || check_pass(u, arg[2], c->sesskey)))
    {
      if (u->passwd[0] == '$')
	{
	  put_string(c, "-This is external account, password change not supported\n");
	  return 1;
	}
      modify_passwd(u, arg[3]);
      do_log("PASSWD %s from %s succeeded", arg[1], conn_name(c));
      write_users();
      return 0;
    }

  do_log("PASSWD %s from %s failed", arg[1], conn_name(c));
  syslog(LOG_WARNING, "PASSWD %s from %s failed\n", arg[1], conn_name(c));
  waiting(c, SUA_RETRY_TIMEOUT);
  GO(WAIT);
  return 1;
}

static int
cmd_who(struct connection *c)
{
  c->tmp = 0;
  waiting(c, 0);
  GO(WHO);
  return 1;
}

static int
cmd_nop(struct connection *c)
{
  return 0;
}

static struct command commands[] = {
  { "quit", 0, 0, cmd_quit },
  { "adduser", 2, AU_ROOT, cmd_add },
  { "deluser", 1, AU_ROOT, cmd_del },
  { "passwd", 3, AU_LOCAL, cmd_passwd },
  { "salt", 1, 0, cmd_salt },
  { "auth", 2, 0, cmd_auth },
  { "login", 4, 0, cmd_login },
  { "clear", 1, 0, cmd_clear },
  { "master", 0, AU_PORT, cmd_master },
  { "who", 0, 0, cmd_who },
  { "nop", 0, 0, cmd_nop }
};

static void
execute_command(struct connection *c, char *cmd)
{
  int argc = 0;
  int i;

  DBG((">> %s\n", cmd));
  for(;;)
    {
      while (*cmd == ' ' || *cmd == '\t')
	cmd++;
      if (!*cmd)
	break;
      if (argc >= MAXARGS)
	{
	  put_string(c, "-Argument list too long\n");
	  return;
	}
      arg[argc++] = cmd;
      while (*cmd && *cmd != ' ' && *cmd != '\t')
	cmd++;
      if (!*cmd)
	break;
      *cmd++ = 0;
    }

  if (!argc--)
    return;

  for(i=0; i < sizeof(commands) / sizeof(struct command); i++)
    if (!strcasecmp(commands[i].name, arg[0]))
      {
#ifdef DEBUG
	int master = 1;
#else
	int master = ntohs(c->port) == SUA_CLIENT_MASTER_PORT;
#endif

	if (commands[i].nargs > argc)
	  {
	    put_string(c, "-Too few arguments\n");
	    return;
	  }
	if (commands[i].nargs < argc)
	  {
	    put_string(c, "-Too many arguments\n");
	    return;
	  }
	if (commands[i].auth & AU_LOCAL && !c->islocal)
	  {
	    put_string(c, "-Local connection required\n");
	    return;
	  }
	if (commands[i].auth & AU_PORT && !master)
	  {
	    put_string(c, "-You need to be a master to do this\n");
	    return;
	  }
	if (commands[i].auth & AU_ROOT && c->local_uid)
	  {
	    put_string(c, "-You must be root to do this\n");
	    return;
	  }
	if (!commands[i].cmd(c))
	  put_string(c, "+OK\n");
	return;
      }
  put_string(c, "-Unknown command\n");
}

/* States */

static void
do_drop(struct connection *c)
{
  drop_conn(c, 0);
}

static void
do_wait(struct connection *c)
{
  put_string(c, "-Invalid username or password\n");
  GO(COMMAND);
}

static void
do_hello(struct connection *c)
{
  if (!c->islocal && !match_client(c->client.s_addr))
    {
      syslog(LOG_NOTICE, "Unauthorized access from %s", conn_name(c));
      dropping(c, "-You are not allowed to talk to me!\n");
      return;
    }
  put_string(c, "+SUAS " SUA_VERSION " Ready. Session key is <");
  generate_random(c->sesskey, SUA_SESSION_BIN_KEY_SIZE);
  put_string(c, c->sesskey);
  put_string(c, ">\n");
  GO(COMMAND);
}

static void
do_command(struct connection *c)
{
  char *k;

  if (k = get_string(c))
    execute_command(c, k);
}

static void
do_who(struct connection *c)
{
  int i = c->tmp;
  struct uid_rec *u = first_uid_rec;
  struct logged_in *l;
  struct user_info *q;
  char buf[32];
  int cnt = 0;

  while (i && u)
    u = u->next, i--;
  if (!u)
    {
      put_string(c, "+DONE\n");
      GO(COMMAND);
    }
  else
    {
      c->tmp++;
      l = u->in;
      while (l)
	{
	  cnt++;
	  l = l->next;
	}
      q = find_user_uid(u->uid);
      sprintf(buf, "+%08x\t%s\t%d\n", u->addr, q ? q->name : "Someone Else", cnt);
      put_string(c, buf);
    }
}

/* Connection engine */

static struct connection *
new_conn(int sock, int state, int wait)
{
  struct connection *c;

  if (num_connects >= MAX_CONNECTIONS)
    return NULL;
  num_connects ++;

  c = xmalloc(sizeof(struct connection));

  c->next = first_connection;
  c->prev = NULL;
  if (first_connection)
    first_connection->prev = c;
  first_connection = c;

  c->id = ++conn_id;
  c->sock = sock;
  c->state = state;
  c->waiting = wait;
  c->timeout = 0;
  c->ismaster = 0;
  c->rbend = c->rbptr = c->rbuf;
  c->wbend = c->wbptr = c->wbuf;

  DBG(("%d: New connection, state=%d, wait=%d, sock=%d\n", c->id, c->state, c->waiting, c->sock));

  return c;
}

static void
drop_conn(struct connection *c, int why)
{
  if (why)
    syslog(LOG_INFO, "Connection to %s closed (%s)",
	   conn_name(c), strerror(why));
#ifndef NOT_HUNTING_DEADLY_BUGS
  else
    syslog(LOG_INFO, "Connection to %s closed normally",
	   conn_name(c));
#endif

  DBG(("%d: Dropped\n", c->id));

  if (c->ismaster)
    dropping_master(c);
  num_connects--;
  close(c->sock);
  if (c->next)
    c->next->prev = c->prev;
  if (c->prev)
    c->prev->next = c->next;
  else
    first_connection = c->next;
  free(c);
}

static void
init_tcp_conn(void)
{
  struct sockaddr_in addr;
  int one = 1;

  sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sock < 0)
    {
      syslog(LOG_ERR, "socket: %m");
      exit(1);
    }

  if (fcntl(sock, F_SETFL, O_NONBLOCK))
    {
      syslog(LOG_ERR, "fcntl: %m");
      exit(1);
    }

  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) < 0)
    {
      syslog(LOG_ERR, "setsockopt: %m");
      exit(1);
    }

  addr.sin_family = AF_INET;
  addr.sin_port = htons(SUA_SERVER_PORT);
  addr.sin_addr.s_addr = INADDR_ANY;

  if (bind(sock, (struct sockaddr *) &addr, sizeof(addr)))
    {
      syslog(LOG_ERR, "bind: %m");
      exit(1);
    }

  if (listen(sock, 4))
    {
      syslog(LOG_ERR, "listen: %m");
      exit(1);
    }

  new_conn(sock, STATE_ACCEPT, WAIT_READ);
}

static void
init_unix_conn(void)
{
  struct sockaddr_un addr;

  sock = socket(AF_UNIX, SOCK_STREAM, 0);
  if (sock < 0)
    {
      syslog(LOG_ERR, "socket: %m");
      exit(1);
    }

  if (fcntl(sock, F_SETFL, O_NONBLOCK))
    {
      syslog(LOG_ERR, "fcntl: %m");
      exit(1);
    }

  addr.sun_family = AF_UNIX;
  strcpy(addr.sun_path, SUA_DAEMON_SOCKET);

  unlink(SUA_DAEMON_SOCKET);
  if (bind(sock, (struct sockaddr *) &addr, sizeof(addr)))
    {
      syslog(LOG_ERR, "bind: %m");
      exit(1);
    }

#ifdef USE_UNIX_SOCK_AUTH
  {
    int one = 1;
    if (setsockopt(sock, SOL_SOCKET, SO_PASSCRED, &one, sizeof(one)) < 0)
      {
	syslog(LOG_ERR, "setsockopt: %m");
	exit(1);
      }
  }
#else
  if (chmod(SUA_DAEMON_SOCKET, 0600) < 0)
    {
      syslog(LOG_ERR, "%s: chown: %m", SUA_DAEMON_SOCKET);
      exit(1);
    }
#endif

  if (listen(sock, 4))
    {
      syslog(LOG_ERR, "listen: %m");
      exit(1);
    }

  new_conn(sock, STATE_ACCEPT, WAIT_READ);
}

#ifdef USE_UNIX_SOCK_AUTH

static int
get_cred(int s)
{
  struct msghdr m;
  char cbuf[1024];
  struct cmsghdr *c;

  m.msg_name = NULL;
  m.msg_namelen = 0;
  m.msg_iov = NULL;
  m.msg_iovlen = 0;
  m.msg_control = cbuf;
  m.msg_controllen = sizeof(cbuf);
  m.msg_flags = 0;
  if (recvmsg(s, &m, 0) < 0)
    {
      syslog(LOG_ERR, "get_cred: %m");
      return -1;
    }
  if (m.msg_flags & MSG_CTRUNC)
    {
      syslog(LOG_ERR, "get_cred: cmsg truncated");
      return -1;
    }
  for(c = CMSG_FIRSTHDR(&m); c; c = CMSG_NXTHDR(&m, c))
    if (c->cmsg_level == SOL_SOCKET && c->cmsg_type == SCM_CREDENTIALS)
      {
	struct ucred *u = CMSG_DATA(c);
	DBG(("Got credentials: uid=%d, gid=%d\n", u->uid, u->gid));
	return u->uid;
      }
  syslog(LOG_ERR, "get_cred: credentials not found");
  return -1;
}

#else

#define get_cred(i) 0

#endif

static void schedule(struct connection *);

static void
do_accept(struct connection *c)
{
  int i, j;
  struct sockaddr_in sa;
  struct connection *d;

  j = sizeof(sa);
  while ((i = accept(c->sock, (struct sockaddr *) &sa, &j)) > 0)
    {
      d = new_conn(i, STATE_HELLO, 0);
      if (sa.sin_family == AF_INET)
	{
	  d->client = sa.sin_addr;
	  d->port = sa.sin_port;
	  d->islocal = IN_LOOPBACK(htonl(d->client.s_addr));
	  d->local_uid = -1;
	}
      else
	{
	  d->client.s_addr = INADDR_ANY;
	  d->port = 0;
	  d->islocal = 1;
	  d->local_uid = get_cred(i);
	}
      schedule(d);
    }
  if (i < 0 && errno != EAGAIN)
    syslog(LOG_ERR, "accept: %m");
}

static void (*staters[STATE_MAX])(struct connection *) = {
  do_accept, do_hello, do_command, do_drop, do_wait, do_who
};

static void
schedule(struct connection *c)
{
  DBG2(("%d: executing state %d\n", c->id, c->state));
  if (c->state < 0 || c->state >= STATE_MAX)
    {
      syslog(LOG_CRIT, "Unknown state %d for %s", c->state, conn_name(c));
      drop_conn(c, EDEADLK);
    }
  staters[c->state](c);
  DBG2(("%d: waiting for %d\n", c->id, c->waiting));
}

static void
run_timers(time_t now)
{
  struct connection *c, *d;

  for(c = first_connection; c; c = d)
    {
      d = c->next;
      if (c->timeout && c->timeout <= now)
	{
	  DBG2(("%d: Timeout\n", c->id));
	  if (c->waiting & WAIT_TIMER)
	    {
	      c->waiting = 0;
	      c->timeout = ~0;
	      schedule(c);
	    }
	  else
	    drop_conn(c, ETIMEDOUT);
	}
    }
}

static void
single_step(void)
{
  fd_set fdr, fdw;
  struct connection *c, *d;
  ulg tout = ~0;
  struct timeval tv, *ptv;
  time_t now;
  int i, max;

  FD_ZERO(&fdr);
  FD_ZERO(&fdw);
  max = 0;
  for(c = first_connection; c; c = c->next)
    {
      if (c->waiting)
	{
	  if (c->sock > max)
	    max = c->sock;
	  if (c->waiting & WAIT_READ)
	    FD_SET(c->sock, &fdr);
	  if (c->waiting & WAIT_WRITE)
	    FD_SET(c->sock, &fdw);
	}
      if (c->timeout && tout > c->timeout)
	tout = c->timeout;
    }

  if (tout == ~0)
    ptv = NULL;
  else
    {
      now = time(NULL);
      if (now >= tout)
	{
	  run_timers(now);
	  return;
	}
      tv.tv_sec = tout - now;
      tv.tv_usec = 0;
      ptv = &tv;
    }

  i = select(max+1, &fdr, &fdw, NULL, ptv);
  DBG2(("Selected %d\n", i));
  if (i < 0)
    syslog(LOG_ERR, "select: %m");
  else if (!i)
    run_timers(time(NULL));
  else
    {
      for(c = first_connection; c; c = d)
	{
	  d = c->next;
	  if (FD_ISSET(c->sock, &fdr))
	    {
	      if (c->state == STATE_ACCEPT)
		{
		  schedule(c);
		  continue;
		}
	      i = read(c->sock, c->rbend, c->rbuf + sizeof(c->rbuf) - c->rbend);
	      DBG2(("%d: Read %d bytes\n", c->id, i));
	      if (i < 0)
		{
		  drop_conn(c, errno);
		  continue;
		}
	      if (!i)
		{
		  drop_conn(c, 0);
		  continue;
		}
	      if (i > 0)
		{
		  while (i--)
		    if (*c->rbend++ == '\n')
		      c->waiting &= ~WAIT_READ;
		  if (c->rbend == c->rbuf + sizeof(c->rbuf))
		    {
		      drop_conn(c, E2BIG);
		      continue;
		    }
		}
	    }
	  if (FD_ISSET(c->sock, &fdw))
	    {
	      i = write(c->sock, c->wbptr, c->wbend - c->wbptr);
	      DBG2(("%d: Wrote %d bytes\n", c->id, i));
	      if (i < 0)
		{
		  drop_conn(c, errno);
		  continue;
		}
	      c->wbptr += i;
	      if (c->wbptr == c->wbend)
		{
		  c->wbptr = c->wbend = c->wbuf;
		  c->waiting &= ~WAIT_WRITE;
		}
	    }
	  if (!c->waiting)
	    schedule(c);
	}
    }
}

static void
sig_ignore(int unused)
{
}

int
main(int argc, char **argv)
{
#ifndef DEBUG
  int i = fork();
  if (i < 0)
    die("fork: %m");
  if (i > 0)
    return 0;
#endif

  openlog(program_name, LOG_NDELAY | LOG_PID, LOG_AUTHPRIV);
  setsig(SIGPIPE, sig_ignore);
  log_open();
  do_log("Daemon started");
  syslog(LOG_NOTICE, "Daemon started");
  read_domains();
  read_users();
  init_tcp_conn();
  init_unix_conn();

  init_nfs();
  cmd_nfs(0, UICCLEAR, 0);

  for(;;)
    single_step();
}
