/*
 *	Simple User Authentication System -- Library Functions
 *
 *	(c) 1996--1998 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License. See file COPYING in any of the GNU packages.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/signal.h>

#include "suas.h"
#include "md5.h"

void
die(char *msg, ...)
{
  va_list args;

  va_start(args, msg);
  fprintf(stderr, "%s: ", program_name);
  vfprintf(stderr, msg, args);
  fputs("!\n", stderr);
  exit(1);
}

void *
xmalloc(ulg size)
{
  void *k = malloc(size);
  if (!k)
    die("Out of memory");
  return k;
}

static char b64c[64] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";

void
encode_base64(byte *from, char *to, int len)
{
  while (len)
    {
      ulg k;
      int c;

      c = (len < 3) ? len : 3;
      k = 0;
      len -= c;
      while (c--)
	k = (k << 8) | *from++;
      *to++ = b64c[ (k >> 18) & 0x3f ];
      *to++ = b64c[ (k >> 12) & 0x3f ];
      *to++ = b64c[ (k >> 6) & 0x3f ];
      *to++ = b64c[ k & 0x3f ];
    }
  *to++ = 0;
}

void
encrypt_first(char *passwd, char *dest, char *salt)
{
  struct MD5Context mc;
  byte final[16];

  if (salt[0] == '$')
    strcpy(dest, passwd);
  else
    {
      MD5Init(&mc);
      MD5Update(&mc, passwd, strlen(passwd));
      MD5Final(final, &mc);
      encode_base64(final, dest, SUA_USER_BIN_KEY_SIZE);
    }
}

void
encrypt_local(char *passwd, char *dest, char *salt)
{
  struct MD5Context mc;
  byte final[16];
  int i;

  if (salt[0] == '$')
    {
#ifdef USE_LOCAL_PASSWORDS
      if (salt[1] == 'L')
	strcpy(dest, pw_encrypt(passwd, salt+2));
      else
#endif
	strcpy(dest, "?");
    }
  else
    {
      int sl = strlen(salt);
      MD5Init(&mc);
      for(i=0; i<64; i++)
	{
	  if (!(i % 3))
	    MD5Update(&mc, salt, sl);
	  if (i % 5)
	    MD5Update(&mc, passwd, strlen(passwd));
	}
      MD5Final(final, &mc);
      encode_base64(final, dest, SUA_USER_BIN_KEY_SIZE);
    }
}

void
encrypt_net(char *local, char *dest, char *session_key)
{
  struct MD5Context mc;
  byte final[16];
  int i;
  int ll = strlen(local);

  MD5Init(&mc);
  for (i=0; i<64; i++)
    {
      if (!(i % 3))
	MD5Update(&mc, session_key, SUA_SESSION_KEY_SIZE);
      if (i % 5)
	MD5Update(&mc, local, ll);
    }
  MD5Final(final, &mc);
  encode_base64(final, dest, SUA_NET_BIN_KEY_SIZE);
}

static void
init_pool(byte *pool)
{
  int f;

  f = open("/dev/urandom", O_RDONLY);
  if (f > 0)
    {
      read(f, pool, 64);
      close(f);
    }
  else
    die("Unable to open /dev/urandom: %m");
}

static char *
update_pool(void)
{
  static int random_init = 0;
  static byte pool[64];
  ulg now = time(NULL);
  int i, z;

  if (!random_init)
    {
      init_pool(pool);
      random_init = 1;
    }

  z = random() + now;
  for(i=0; i<64; i++)
    {
      pool[i] ^= (pool[(i+17)& 63] + 7*pool[(pool[i] & 63) ^ 23] - 13*pool[(z >> 17) & 63])
	+ (z >> 7);
      z = z * 259309 + 151;
    }
  return pool;
}

void
generate_random(char *dest, ulg size)	/* size < 16 ! */
{
  byte buf[16];
  struct MD5Context mc;
  char *pool;

  pool = update_pool();
  MD5Init(&mc);
  MD5Update(&mc, pool, 16);
  MD5Final(buf, &mc);
  encode_base64(buf, dest, size);
}

void
setsig(int sig, void (*func)(int))
{
  struct sigaction a;

  bzero(&a, sizeof(a));
  a.sa_handler = func;
  a.sa_flags = SA_RESTART;
  if (sigaction(sig, &a, NULL))
    {
      syslog(LOG_ERR, "Setup of signal %d failed", sig);
      exit(1);
    }
}
