/*
 *	Simple User Authentication System -- Client Master
 *
 *	(c) 1996 -- 1998 Martin Mares, <mj@atrey.karlin.mff.cuni.cz>
 *	(c) 1999 Martin Hinner, <martin@tdp.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License. See file COPYING in any of the GNU packages.
 */

#include <stdio.h>			/* Includes, includes ... and well, some includes ... */
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <stdarg.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <syslog.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/wait.h>
#include <pwd.h>
#include <errno.h>
#include <sys/un.h>
#include <sys/time.h>

#include "suas.h"
#include "master.h"

char program_name[] = "suamaster";

/*
 *	Altering Program Title
 */

static char *title_start, *title_end;

static void
title_init(char **argv, char **envp)
{
  title_start = argv[0];
  while (*envp)
    envp++;
  title_end = envp[-1] + strlen(envp[-1]);
}

void
title_set(char *str, ...)
{
  static char buf[256];
  va_list args;
  char *c, *d;

  va_start(args, str);
  vsprintf(buf, str, args);
  va_end(args);
  DBG(("[%s]\n", buf));
  c = buf;
  d = title_start;
  while (*c && d < title_end)
    *d++ = *c++;
  while (d < title_end)
    *d++ = 0;
}

/*
 *    	Auth Server Connections
 */

struct sconn {
  struct sconn *next;
  struct sockaddr_in sad;
  int fd;				/* -1 if connection lost */
  char inbuf[256];
  char session_key[80];
  char *result;
  char name[1];
};

static struct sconn *first_conn;
static char *primary_name;

static void
sconn_lost(struct sconn *c)
{
  DBG(("Lost connection to %s", c->name));
  if (c->fd >= 0)
    {
      syslog(LOG_ERR, "Connection to %s lost (%m). Wait the rain", c->name);
      close(c->fd);
      c->fd = -1;
    }
}

static int				/* <0=err, 0=OK, 1=OK+result */
reply(struct sconn *c)
{
  int z, l;
  char *b;

  b = c->inbuf;
  l = sizeof(c->inbuf);
  for(;;)
    {
      z = read(c->fd, b, l);
      if (z < 0 && errno == EAGAIN)
	continue;
      if (z == l)
	{
	  syslog(LOG_ERR, "Server reply too long. Confused!");
	  goto lost;
	}
      if (!z)
	{
	lost:
#ifndef NOT_HUNTING_DEADLY_BUGS
	  syslog(LOG_INFO, "Read returned an error: %m");
#endif
	  sconn_lost(c);
	  return -2;
	}
      if (b[z-1] == '\r' || b[z-1] == '\n')
	break;
      b += z;
      l -= z;
    }
  if (b[z-1] == '\n')
    b[--z] = 0;
  if (z && b[z-1] == '\r')
    b[--z] = 0;
  DBG(("Got reply /%s/ from %s\n", c->inbuf, c->name));

  if (c->inbuf[0] == '-')
    return -1;

  b = strchr(c->inbuf, '<');
  if (b)
    {
      c->result = b+1;
      if (b = strchr(c->inbuf, '>'))
	{
	  *b = 0;
	  return 1;
	}
    }
  c->result = NULL;
  return 0;
}

static int
cmd(struct sconn *c, char *cmd, ...)
{
  va_list args;
  char buf[256], *bp;
  int flag = 0;
  int l, r;

  va_start(args, cmd);
  if (cmd[0] == '-')
    {
      cmd++;
      flag = 1;
    }
  l = vsprintf(buf, cmd, args);
  va_end(args);
  DBG(("Sending /%s/ to %s\n", buf, c->name));
  buf[l++] = '\n';
  bp = buf;
  while (l)
    {
      r = write(c->fd, bp, l);
      if (r < 0 && errno == EAGAIN)
	continue;
      if (r <= 0)
	{
#ifndef NOT_HUNTING_DEADLY_BUGS
	  syslog(LOG_INFO, "Write returned an error: %m");
#endif
	  sconn_lost(c);
	  return -2;
	}
      bp += r;
      l -= r;
    }
  return reply(c);
}

static void
sockfail(char *e)
{
  syslog(LOG_ERR, "%s: %m", e);
}

static int
sconn_connect(struct sconn *c)
{
  int sock;
  struct sockaddr_in local;
  int one = 1;

  title_set("sua: Connecting to %s", c->name);

  sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sock < 0)
    {
      sockfail("socket");
      return 0;
    }

  local.sin_family = AF_INET;
  local.sin_port = htons(SUA_CLIENT_MASTER_PORT);
  local.sin_addr.s_addr = INADDR_ANY;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)))
    {
      sockfail("setsockopt");
    lost:
      sconn_lost(c);
      return 0;
    }
  if (bind(sock, (struct sockaddr *) &local, sizeof(local)))
    {
      sockfail("bind");
      return 0;
    }

  if (connect(sock, (struct sockaddr *) &c->sad, sizeof(c->sad)))
    {
      sockfail("connect");
      return 0;
    }

  c->fd = sock;
  if (reply(c) <= 0)
    goto lost;
  strcpy(c->session_key, c->result);
  if (cmd(c, "master") < 0)
    goto lost;
  return 1;
}

static struct sconn *
sconn_new(char *name)
{
  struct sconn *c = xmalloc(sizeof(struct sconn) + strlen(name));
  struct hostent *he;

  strcpy(c->name, name);
  if (!(he = gethostbyname(name)))
    {
      syslog(LOG_ERR, "Cannot resolve server name: %m");
      free(c);
      return NULL;
    }
  c->sad.sin_family = AF_INET;
  c->sad.sin_port = htons(SUA_SERVER_PORT);
  memcpy(&c->sad.sin_addr.s_addr, he->h_addr, he->h_length);
  c->fd = -1;

  c->next = first_conn;
  first_conn = c;

  return c;
}

static struct sconn *
sconn_find(char *name)
{
  struct sconn *c;

  for(c=first_conn; c; c=c->next)
    if (!strcmp(c->name, name))
      {
	if (c->fd < 0 && !sconn_connect(c))
	  return NULL;
	return c;
      }

  c = sconn_new(name);
  if (!c || !sconn_connect(c))
    return NULL;

  return c;
}

static void
connect_to_primary(char *name)
{
  struct sconn *c;

  primary_name = xmalloc(strlen(name) + 1);
  strcpy(primary_name, name);
  sconn_find(name);
  if (!(c = first_conn))
    {
      syslog(LOG_ERR, "Cannot resolve primary authentication server name, giving up!");
      sleep(60);
      exit(1);
    }
  if (c->fd < 0)
    syslog(LOG_ERR, "Primary authentication server not available, try praying");
}

static void
sconn_refresh_all(void)
{
  struct sconn *c;

  DBG(("Time Tick\n"));
  for(c=first_conn; c; c=c->next)
    cmd(c, "nop");
}

/*
 *	Getties.
 */

struct getty {
  int pid;
  struct sconn *server;
  int session;
  char name[8];
};

static struct getty *getties;
static int ngetties;

#ifndef SUA_STANDALONE
static void
init_getties(int n, char **p)
{
  int i;

  ngetties = n;
  getties = xmalloc(sizeof(struct getty) * ngetties);
  bzero(getties, sizeof(struct getty) * ngetties);
  for(i=0; i<n; i++)
    {
      struct getty *g = getties + i;
      sprintf(g->name, "tty%s", p[i]);
    }
}
#endif /* SUA_STANDALONE */

static void
finish_session(struct sconn *c, int sess)
{
  DBG(("Finish session %d at %s\n", sess, c->name));
  if (c->fd >= 0)			/* If not connected, can safely forget the clear */
    cmd(c, "clear %d", sess);
}

/*
 *	Communication with local clients.
 */

static int unix_sock;

struct client {
  struct client *next;
  int sock;
  char buf[256];
  int bpos, towrite;
  int session;
  struct sconn *server;
};

static struct client *first_client;

static void unfail(char *) NO_RET;
static void cl_write(struct client *, char *);

static void
unfail(char *c)
{
  syslog(LOG_ERR, "%s: %m", c);
  exit(1);
}

struct sockaddr_un master_sa = { AF_UNIX, SUA_MASTER_SOCKET };

static void
create_unix_socket(void)
{
  unix_sock = socket(AF_UNIX, SOCK_STREAM, 0);
  if (unix_sock < 0)
    unfail("socket");
  unlink(master_sa.sun_path);
  if (bind(unix_sock, (struct sockaddr *) &master_sa, sizeof(master_sa)) < 0)
    unfail("bind");
  if (listen(unix_sock, 4) < 0)
    unfail("listen");
  if (fcntl(unix_sock, F_SETFL, O_NONBLOCK) < 0)
    unfail("fcntl");
}

static int
free_session_p(int s)
{
  struct client *c;

  for(c=first_client; c; c=c->next)
    if (c->session == s)
      return 0;
  return 1;
}

static void
get_free_session(struct client *c)
{
  static int sess_counter;

  for(;;)
    {
      int s = sess_counter + 100;
      if (free_session_p(s))
	{
	  c->session = s;
	  return;
	}
      sess_counter = (sess_counter + 1) % 256;
    }
}

static void
close_conn(struct client *c)
{
  struct client **d;

  DBG(("Closing connection %d\n", c->sock));
  close(c->sock);
  if (c->server)
    finish_session(c->server, c->session);
  d = &first_client;
  while (*d)
    {
      if (*d == c)
	{
	  *d = c->next;
	  free(c);
	  return;
	}
      d = &(*d)->next;
    }
  syslog(LOG_ERR, "Client not found for socket %d", c->sock);
}

static char *
handle_req_remote(struct client *c, char *serv, struct passwd *pw, char *pass, struct sconn **SC)
{
  struct sconn *sc;
  char salt[SUA_USER_SALT_SIZE+1];
  byte uk[SUA_USER_KEY_SIZE+1], nk[SUA_NET_KEY_SIZE+1], rk[SUA_USER_KEY_SIZE+1];
  char *user = pw->pw_name;

  DBG(("Remote auth at %s\n", serv));

  if (!(sc = sconn_find(serv)))
    return "-Authentication server not available";
  *SC = sc;

  title_set("sua: user %s", user);
  if (cmd(sc, "-salt %s", user) <= 0)
    return sc->inbuf;
  strcpy(salt, sc->result);
  encrypt_first(pass, rk, salt);
  encrypt_local(rk, uk, salt);
  encrypt_net(uk, nk, sc->session_key);
  
  if (cmd(sc, "login %s %s %d %s", user, nk, c->session, sc->session_key) < 0)
    return sc->inbuf;

  return "+Logged in";
}

static char *
handle_req_local(struct client *c, struct passwd *pw, char *pass)
{
#ifdef LOCAL_AUTH
  struct spwd *sp;
  char *p = pw->pw_passwd;

  DBG(("Local auth...\n"));

  if (*p == '*' && !p[1])
    return "-Account disabled";
  endspent();
  if (sp = getspnam(pw->pw_name))
    p = sp->sp_pwdp;
  if (*p == '@' || (*p && strlen(p) < 3))
    return NULL;
  if (*p && strcmp(p, pw_encrypt(pass, p)))
    return "-Invalid password";
  return "+Logged in locally";
#else
  return NULL;
#endif
}

static void
handle_request(struct client *c)
{
  char *r = c->buf;
  char *sep1, *sep2, *serv, *msg;
  int s = -1;
  struct passwd *pw;
  struct getty *g = NULL;
  struct sconn *sc = NULL;

  DBG(("Request `%s'\n", r));
  c->session = -1;			/* Don't accept more requests */
  if (! (sep1 = strchr(r, '\001')))
    {
      cl_write(c, "-Invalid request");
      return;
    }
  *sep1++ = 0;
  if (sep2 = strchr(sep1, '\001'))
    {
      *sep2++ = 0;
      sscanf(sep2, "%d", &s);
    }

  if (s >= 1 && s <= ngetties)
    {
      g = getties + s - 1;
      if (g->server)
	{
	  syslog(LOG_ERR, "Double lock on channel %d", s);
	  cl_write(c, "-Double lock");
	  return;
	}
      c->session = g->session = s;
    }
  else
    get_free_session(c);

  endpwent();
  pw = getpwnam(r);
  if (!pw)
    {
      sleep(1);
      cl_write(c, "-No such user");
      return;
    }
  if (pw->pw_passwd[0] == '@' && strlen(pw->pw_passwd) > 3) /* Server name stored in passwd */
    serv = pw->pw_passwd + 1;
  else
    serv = primary_name;
  if (!(msg = handle_req_local(c, pw, sep1)))
    msg = handle_req_remote(c, serv, pw, sep1, &sc);
  if (*msg != '+')
    {
      if (*msg != '-')
	msg = "-Authentication failed";
      cl_write(c, msg);
      return;
    }

  if (s < 0)
    c->server = sc;
  else
    g->server = sc;
  cl_write(c, msg);

  title_set("suamaster");
  syslog(LOG_INFO, "%s logged in, session=%d", pw->pw_name, ((s > 0) ? s : 0));
}

static void
handle_conn(struct client *c)
{
  int l;
  char *x;

restart:
  if (c->towrite)
    {
      l = write(c->sock, c->buf + c->bpos, c->towrite - c->bpos);
      if (l > 0)
	{
	  c->bpos += l;
	  if (c->bpos >= c->towrite)
	    {
	      c->towrite = c->bpos = 0;
	      return;
	    }
	  goto restart;
	}
    }
  else
    {
      l = read(c->sock, c->buf + c->bpos, sizeof(c->buf) - c->bpos - 1);
      if (l > 0)
	{
	  c->buf[c->bpos + l] = 0;
	  x = strchr(c->buf + c->bpos, '\n');
	  if (x)
	    {
	      if (!c->session)
		{
		  *x = 0;
		  handle_request(c);
		}
	      c->bpos = 0;
	      goto restart;
	    }
	  else
	    {
	      c->bpos += l;
	      if (c->bpos >= sizeof(c->buf) - 1)
		l = -1;
	    }
	}
    }
  if (!l || errno != EAGAIN)
    close_conn(c);
}

static void
cl_write(struct client *c, char *msg)
{
  c->bpos = 0;
  c->towrite = strlen(msg);
  strcpy(c->buf, msg);
  c->buf[c->towrite++] = '\n';
  handle_conn(c);
}

static void
handle_incoming(void)
{
  int s, al;
  struct sockaddr sa;
  struct client *c;

  for(;;)
    {
      al = sizeof(sa);
      s = accept(unix_sock, &sa, &al);
      if (s < 0)
	break;
      fcntl(s, F_SETFL, O_NONBLOCK);
      DBG(("Incoming connection %d\n", s));
      c = xmalloc(sizeof(struct client));
      c->next = first_client;
      first_client = c;
      c->sock = s;
      c->bpos = c->towrite = c->session = 0;
      c->server = NULL;
      cl_write(c, "+Welcome in hell");
    }
}

static void
handle_files(void)
{
  fd_set rd, wr;
  int hi;
  struct client *c, *d;

  FD_ZERO(&rd);
  FD_ZERO(&wr);
restart:
  FD_SET(hi = unix_sock, &rd);
  for(c=first_client; c; c=c->next)
    {
      if (c->towrite < 0)
	FD_SET(c->sock, &wr);
      else
	FD_SET(c->sock, &rd);
      if (c->sock > hi)
	hi = c->sock;
    }

  hi = select(hi+1, &rd, &wr, NULL, NULL);
  if (hi > 0)
    {
      if (FD_ISSET(unix_sock, &rd))
	handle_incoming();
      for(c=first_client; c; c=d)
	{
	  d = c->next;
	  if (FD_ISSET(c->sock, &rd) || FD_ISSET(c->sock, &wr))
	    handle_conn(c);
	}
      goto restart;
    }
}

/*
 *	Wanted: Processes. Dead or Alive.
 */

#ifndef SUA_STANDALONE
static void
start_getty(int i, struct getty *g)
{
  struct client *c;

  close(0);
  close(1);
  close(2);
  close(unix_sock);
  for(c=first_client; c; c=c->next)
    close(c->sock);
  title_set("getty");
  run_getty(i, g->name);		/* syslog is still open... */
  exit(23);
}

static void
restart(int pid)
{
  struct getty *g;
  int i, p;

  for(i=0; i<ngetties; i++)
    {
      g = getties + i;
      if (!g->pid || g->pid == pid)
	{
	  g->pid = 0;
	  if (g->server)
	    {
	      finish_session(g->server, g->session);
	      g->server = NULL;
	    }
	  p = fork();
	  if (p < 0)
	    {
	      syslog(LOG_ERR, "fork: %m");
	      continue;
	    }
	  if (!p)
	    start_getty(i+1, g);
	  g->pid = p;
	}
    }
}
#endif /* SUA_STANDALONE */

static int alarm_flag;

static void
alarm_handler(int unused)
{
  alarm_flag = 1;
}

static void
child_handler(int unused)
{
}

/*
 *	Main Loop.
 */

int
main(int argc, char **argv, char **envp)
{
  openlog(argv[0], LOG_NDELAY | LOG_PID, LOG_DAEMON);

#ifdef SUA_STANDALONE
  syslog(LOG_INFO, "suamaster standalone version " SUA_VERSION " starting");
#else
  syslog(LOG_INFO, "suamaster with getties version " SUA_VERSION " starting");
#endif

#ifndef DEBUG
  close(0);
  close(1);
  close(2);
  setsid();
#endif

#ifdef SUA_STANDALONE
  if (argc != 2)
#else
  if (argc < 3)
#endif
    {
      syslog(LOG_ERR, "Invalid arguments");
      fprintf(stderr,"suamaster: Invalid arguments\n");
      exit(1);
    }

#ifdef SUA_STANDALONE
  if (fork())                   /* Daemonize... */
   exit(0);
#endif

#ifndef SUA_STANDALONE
  init_getties(argc - 2, argv + 2);
#endif

  title_init(argv, envp);
  connect_to_primary(argv[1]);
  create_unix_socket();
#ifndef SUA_STANDALONE
  restart(0);
#endif
  setsig(SIGALRM, alarm_handler);
  setsig(SIGCHLD, child_handler);
  alarm(SUA_REFRESH);

  for(;;)
    {
#ifndef SUA_STANDALONE
      int status, pid;

      while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
	{
	  DBG(("Died PID %d, status=%x\n", pid, status));
	  restart(pid);
	}
#endif

      if (alarm_flag)
	{
	  alarm_flag = 0;
	  sconn_refresh_all();
	  alarm(SUA_REFRESH);
	}

      handle_files();
      title_set("suamaster");
    }
}
